﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
// Some namespace mappings
using MF = SharpDX.MediaFoundation;

namespace Testing.SimpleMFPlayer
{
    class MFSessionEventListener : MF.IAsyncCallback
    {
        private SynchronizationContext m_syncContext;
        private MF.MediaSession m_session;

        /// <summary>
        /// Raised when the media session has closed.
        /// </summary>
        public event EventHandler SessionClosed;

        /// <summary>
        /// Raised when the media topology is ready.
        /// </summary>
        public event EventHandler TopologyReady;

        /// <summary>
        /// Raises when the end of a video is reached.
        /// </summary>
        public event EventHandler EndOfPresentation;

        /// <summary>
        /// Initializes a new instance of the <see cref="MFSessionEventListener"/> class.
        /// </summary>
        /// <param name="session">The session to listen to.</param>
        public MFSessionEventListener(MF.MediaSession session)
        {
            m_syncContext = SynchronizationContext.Current;
            m_session = session;

            // Start listening on the session object
            m_session.BeginGetEvent(this, null);
        }

        /// <summary>
        /// Führt anwendungsspezifische Aufgaben aus, die mit dem Freigeben, Zurückgeben oder Zurücksetzen von nicht verwalteten Ressourcen zusammenhängen.
        /// </summary>
        public void Dispose()
        {
            if (this.Shadow != null)
            {
                this.Shadow.Dispose();
                this.Shadow = null;
            }
        }

        /// <summary>
        ///   <p><strong>Applies to: </strong>desktop apps | Metro style apps</p><p> </p><p>Called when an asynchronous operation is completed.</p>
        /// </summary>
        /// <param name="asyncResultRef"><dd> <p>Pointer to the <strong><see cref="T:SharpDX.MediaFoundation.AsyncResult" /></strong> interface. Pass this reference to the asynchronous <strong>End...</strong> method to complete the asynchronous call.</p> </dd></param>
        /// <!-- Failed to insert some or all of included XML --><include file=".\..\Documentation\CodeComments.xml" path="/comments/comment[@id='IMFAsyncCallback::Invoke']/*" />
        ///   <msdn-id>bb970360</msdn-id>
        ///   <unmanaged>HRESULT IMFAsyncCallback::Invoke([In, Optional] IMFAsyncResult* pAsyncResult)</unmanaged>
        ///   <unmanaged-short>IMFAsyncCallback::Invoke</unmanaged-short>
        /// <remarks>
        ///   <p>Within your implementation of <strong>Invoke</strong>, call the corresponding <strong>End...</strong> method.</p><p>This interface is available on the following platforms if the Windows Media Format 11 SDK redistributable components are installed:</p><ul> <li>Windows?XP with Service Pack?2 (SP2) and later.</li> <li>Windows?XP Media Center Edition?2005 with KB900325 (Windows?XP Media Center Edition?2005) and KB925766 (October 2006 Update Rollup for Windows?XP Media Center Edition) installed.</li> </ul>
        /// </remarks>
        public void Invoke(MF.AsyncResult asyncResultRef)
        {
            MF.MediaEventTypes eventType;

            try
            {
                using (MF.MediaEvent mediaEvent = m_session.EndGetEvent(asyncResultRef))
                {
                    eventType = mediaEvent.TypeInfo;
                    switch (eventType)
                    {
                            // Handle closed session
                        case MF.MediaEventTypes.SessionClosed:

                            // Post corresponding event
                            m_syncContext.Post((arg) =>
                            {
                                if (SessionClosed != null) { SessionClosed(this, EventArgs.Empty); }
                            }, null);

                            // Break this loop here
                            return;

                            // Handle changed toplogy state
                        case MF.MediaEventTypes.SessionTopologyStatus:
                            MF.TopologyStatus state = mediaEvent.Get<MF.TopologyStatus>(MF.EventAttributeKeys.TopologyStatus);
                            if(state == MF.TopologyStatus.Ready)
                            {
                                m_syncContext.Post((arg) =>
                                {
                                    if (TopologyReady != null) { TopologyReady(this, EventArgs.Empty); }
                                }, null);
                            }
                            break;

                        case MF.MediaEventTypes.EndOfPresentation:
                            m_syncContext.Post((arg) =>
                            {
                                if (EndOfPresentation != null) { EndOfPresentation(this, EventArgs.Empty); }
                            }, null);
                            break;
                    }
                }
            }
            catch(Exception ex)
            {
                m_syncContext.Post((arg) =>
                {
                    Application.OnThreadException(ex);
                }, null);
            }

            // Reregister for next event
            m_session.BeginGetEvent(this, null); 
        }

        /// <summary>
        /// Gets the identifier of the work queue on which the callback is dispatched. See remarks.
        /// </summary>
        /// <value>
        /// The work queue identifier.
        /// </value>
        /// <msdn-id>bb970381</msdn-id>
        ///   <unmanaged>HRESULT IMFAsyncCallback::GetParameters([Out] MFASYNC_CALLBACK_FLAGS* pdwFlags,[Out] unsigned int* pdwQueue)</unmanaged>
        ///   <unmanaged-short>IMFAsyncCallback::GetParameters</unmanaged-short>
        /// <exception cref="System.NotImplementedException"></exception>
        /// <remarks>
        ///   <p>This value can specify one of the standard Media Foundation work queues, or a work queue created by the application. For list of standard Media Foundation work queues, see <strong>Work Queue Identifiers</strong>. To create a new work queue, call <strong><see cref="M:SharpDX.MediaFoundation.MediaFactory.AllocateWorkQueue(System.Int32@)" /></strong>. The default value is <strong><see cref="F:SharpDX.MediaFoundation.WorkQueueType.Standard" /></strong>.</p> <p>If the work queue is not compatible with the value returned in <em>pdwFlags</em>, the Media Foundation platform returns <strong><see cref="F:SharpDX.MediaFoundation.ResultCode.InvalidWorkqueue" /></strong> when it tries to dispatch the callback. (See <strong><see cref="M:SharpDX.MediaFoundation.MediaFactory.PutWorkItem(System.Int32,System.IntPtr,SharpDX.ComObject)" /></strong>.)</p>
        /// </remarks>
        public MF.WorkQueueId WorkQueueId
        {
            get { return MF.WorkQueueId.Standard; }
        }

        /// <summary>
        /// Gets or sets the unmanaged shadow callback.
        /// </summary>
        /// <value>
        /// The unmanaged shadow callback.
        /// </value>
        /// <remarks>
        /// This property is set whenever this instance has an unmanaged shadow callback
        /// registered. This callback must be disposed when disposing this instance.
        /// </remarks>
        public IDisposable Shadow
        {
            get;
            set;
        }

        /// <summary>
        /// Gets a flag indicating the behavior of the callback object's <strong><see cref="M:SharpDX.MediaFoundation.IAsyncCallback.Invoke(SharpDX.MediaFoundation.AsyncResult)" /></strong> method. Default behavior should be <see cref="F:SharpDX.MediaFoundation.AsyncCallbackFlags.None" />.
        /// </summary>
        /// <value>
        /// The a flag indicating the behavior of the callback object's <strong><see cref="M:SharpDX.MediaFoundation.IAsyncCallback.Invoke(SharpDX.MediaFoundation.AsyncResult)" /></strong> method.
        /// </value>
        /// <msdn-id>bb970381</msdn-id>
        ///   <unmanaged>HRESULT IMFAsyncCallback::GetParameters([Out] MFASYNC_CALLBACK_FLAGS* pdwFlags,[Out] unsigned int* pdwQueue)</unmanaged>
        ///   <unmanaged-short>IMFAsyncCallback::GetParameters</unmanaged-short>
        /// <exception cref="System.NotImplementedException"></exception>
        public MF.AsyncCallbackFlags Flags
        {
            get { return MF.AsyncCallbackFlags.None; }
        }
    }
}
