﻿using SharpDX;
using SharpDX.MediaFoundation;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

// Some namespace mappings
using MF = SharpDX.MediaFoundation;

namespace Testing.SimpleMFPlayer
{
    public partial class MainWindow : Form
    {
        private string m_currentVideoFile;
        private MF.MediaSession m_mediaSession;
        private MF.SourceResolver m_sourceResolver;
        private MF.MediaSource m_mediaSource;
        private MF.Topology m_topology;
        private MF.PresentationDescriptor m_presentationDescriptor;
        private MF.Activate m_audioRenderer;
        private MF.Activate m_videoRenderer;
        private MFSessionEventListener m_sessionEventHandler;

        /// <summary>
        /// Initializes a new instance of the <see cref="MainWindow"/> class.
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Updates the state of this dialog.
        /// </summary>
        private void UpdateDialogState()
        {
            if(string.IsNullOrEmpty(m_currentVideoFile))
            {
                m_mnuLblVideoFile.Text = "No Video..";
                m_mnuCmdOpenVideo.Enabled = true;
                m_mnuCmdCloseVideo.Enabled = false;
            }
            else
            {
                m_mnuLblVideoFile.Text = m_currentVideoFile;
                m_mnuCmdOpenVideo.Enabled = false;
                m_mnuCmdCloseVideo.Enabled = true;
            }
        }

        /// <summary>
        /// Opens and shows the given video file.
        /// </summary>
        /// <param name="sourceVideoFile">The source video file.</param>
        private void OpenAndShowVideoFile(string sourceVideoFile)
        {
            try
            {
                // Create media session and a corresponding event listener obect for async events
                MF.MediaFactory.CreateMediaSession(null, out m_mediaSession);
                m_sessionEventHandler = new MFSessionEventListener(m_mediaSession);
                m_sessionEventHandler.TopologyReady += OnSessionEventHandlerTopologyReady;
                m_sessionEventHandler.EndOfPresentation += OnSessionEventHandlerEndOfPresentation;

                // Create source object
                MF.MediaFactory.CreateSourceResolver(out m_sourceResolver);
                MF.ObjectType objType = MF.ObjectType.Invalid;
                SharpDX.ComObject objSource = null;
                m_sourceResolver.CreateObjectFromURL(
                    sourceVideoFile,
                    (int)MF.SourceResolverFlags.MediaSource,
                    null,
                    out objType,
                    out objSource);
                m_mediaSource = objSource.QueryInterface<MF.MediaSource>();
                Helper.SafeDispose(ref objSource);

                // Create topology
                MF.MediaFactory.CreateTopology(out m_topology);
                m_mediaSource.CreatePresentationDescriptor(out m_presentationDescriptor);
                int streamDescriptorCount = m_presentationDescriptor.StreamDescriptorCount;
                for (int loop = 0; loop < streamDescriptorCount; loop++)
                {
                    Bool selected = false;
                    StreamDescriptor streamDescriptor;
                    m_presentationDescriptor.GetStreamDescriptorByIndex(loop, out selected, out streamDescriptor);
                    if (selected)
                    {
                        // Create source node
                        MF.TopologyNode sourceNode = null;
                        MF.MediaFactory.CreateTopologyNode(TopologyType.SourceStreamNode, out sourceNode);
                        sourceNode.Set(MF.TopologyNodeAttributeKeys.Source, m_mediaSource);
                        sourceNode.Set(MF.TopologyNodeAttributeKeys.PresentationDescriptor, m_presentationDescriptor);
                        sourceNode.Set(MF.TopologyNodeAttributeKeys.StreamDescriptor, streamDescriptor);

                        // Create output node
                        MF.TopologyNode outputNode = null;
                        MediaTypeHandler mediaTypeHandler = streamDescriptor.MediaTypeHandler;
                        Guid majorType = mediaTypeHandler.MajorType;
                        MF.MediaFactory.CreateTopologyNode(TopologyType.OutputNode, out outputNode);
                        if (MF.MediaTypeGuids.Audio == majorType)
                        {
                            MF.MediaFactory.CreateAudioRendererActivate(out m_audioRenderer);
                            outputNode.Object = m_audioRenderer;
                        }
                        else if (MF.MediaTypeGuids.Video == majorType)
                        {
                            MF.MediaFactory.CreateVideoRendererActivate(
                                m_panVideo.Handle,
                                out m_videoRenderer);
                            outputNode.Object = m_videoRenderer;
                        }

                        // Append nodes to topology
                        m_topology.AddNode(sourceNode);
                        m_topology.AddNode(outputNode);
                        sourceNode.ConnectOutput(0, outputNode, 0);

                        // Clear COM references
                        Helper.SafeDispose(ref sourceNode);
                        Helper.SafeDispose(ref outputNode);
                        Helper.SafeDispose(ref mediaTypeHandler);
                    }

                    // Clear COM references
                    Helper.SafeDispose(ref streamDescriptor);
                }

                // Apply build topology to the session
                m_mediaSession.SetTopology(SessionSetTopologyFlags.None, m_topology);

                // Video opened successfully
                m_currentVideoFile = sourceVideoFile;
            }
            catch(Exception)
            {
                m_currentVideoFile = string.Empty;

                UnloadMFResources();

                throw;
            }

            UpdateDialogState();
        }

        /// <summary>
        /// Unloads all resources of media foundation.
        /// </summary>
        private void UnloadMFResources()
        {
            if(m_mediaSession != null)
            {
                m_mediaSession.Shutdown();
            }

            // Clear all references
            m_currentVideoFile = string.Empty;
            Helper.SafeDispose(ref m_videoRenderer);
            Helper.SafeDispose(ref m_audioRenderer);
            Helper.SafeDispose(ref m_presentationDescriptor);
            Helper.SafeDispose(ref m_topology);
            Helper.SafeDispose(ref m_mediaSource);
            Helper.SafeDispose(ref m_sourceResolver);
            Helper.SafeDispose(ref m_mediaSession);
            m_sessionEventHandler = null;

            this.Invalidate(true);
        }

        /// <summary>
        /// Default load event. 
        /// </summary>
        /// <param name="e"></param>
        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            if (this.DesignMode) { return; }

            MF.MediaFactory.Startup(MF.MediaFactory.Version, 0);

            UpdateDialogState();
        }

        /// <summary>
        /// Called when the current topology is ready for playing.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void OnSessionEventHandlerTopologyReady(object sender, EventArgs e)
        {
            if (m_sessionEventHandler != sender) { return; }

            SharpDX.Win32.Variant v = new SharpDX.Win32.Variant();
            v.Type = SharpDX.Win32.VariantType.Default;
            v.ElementType = SharpDX.Win32.VariantElementType.Empty;

            m_mediaSession.Start(null, v);

            UpdateDialogState();
        }

        /// <summary>
        /// Called when the end of a presentation is reached.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void OnSessionEventHandlerEndOfPresentation(object sender, EventArgs e)
        {
            if (m_sessionEventHandler != sender) { return; }

            UnloadMFResources();

            UpdateDialogState();
        }

        /// <summary>
        /// Called when user clicks exit.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void OnMnuCmdExitClick(object sender, EventArgs e)
        {
            this.Close();
        }

        /// <summary>
        /// Called when user wants to open a video.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void OnMnuCmdOpenVideoClick(object sender, EventArgs e)
        {
            // Query for source file
            if (m_dlgOpenVideoFile.ShowDialog(this) != DialogResult.OK) { return; }
            string sourceVideoFile = m_dlgOpenVideoFile.FileName;

            OpenAndShowVideoFile(sourceVideoFile);
        }

        /// <summary>
        /// Called when user wants to close the currently running video.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void OnMnuCmdCloseVideoClick(object sender, EventArgs e)
        {
            UnloadMFResources();

            this.UpdateDialogState();
        }
    }
}
